//////////////////////////////////////////////////////////////////////////////
//
// Rev          :  $Revision:   1.0  $
// Content      :  Entry (and exit) points for the Linux Driver
// Compile flag :
//
// COPYRIGHT (c) 2003  Matrox Electronic Systems Ltd.
// All Rights Reserved
//
//////////////////////////////////////////////////////////////////////////////
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/kallsyms.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/uaccess.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include "linuxentry.h"
#include "linux-interface.h"

const char *MDrv_DriverName = "irisgtr";
int MDrv_DbgLevel = DBG_CRIT|DBG_ERROR;
int irisgtr_major = 0;
extern unsigned long nbOfBoards;
static struct proc_dir_entry *irisgtr_proc_dir;
static struct proc_dir_entry *irisgtr_volatile_dir;

#define IRISGTR_MAX_CARDS           256

MODULE_LICENSE("Proprietary");
MODULE_DESCRIPTION("MIL irisgtr Driver");
MODULE_AUTHOR("Matrox Electronic Systems Ltd.");

long irisgtr_unlocked_ioctl (struct file *filp, unsigned int cmd, unsigned long arg);
int irisgtr_open (struct inode *inode, struct file *filp);
int irisgtr_mmap(struct file *filp, struct vm_area_struct *vma);
int irisgtr_release (struct inode *inode, struct file *filp);

int  irisgtr_dev_init (struct pci_dev *dev, const struct pci_device_id *id);
void irisgtr_remove (struct pci_dev *dev);
void register_dump (void);
void irisgtr_cleanup (void);
int irisgtr_init (void);

struct file_operations irisgtr_fops =
   {
      .unlocked_ioctl = irisgtr_unlocked_ioctl,
      .open = irisgtr_open,
      .release = irisgtr_release,
      .mmap = irisgtr_mmap
   };

//Structure defining all the IDs that belong to irisgtr
struct pci_device_id irisgtr_ids[] =
   {
      { IRISGTR_VENDID, IRISGTR_DEVICE_ID, IRISGTR_VENDID, IRISGTR_SUBSYS_ID, 0, 0, 0},
      {0,}
   };

//Structure defining the irisgtr driver for hotplug.
struct pci_driver irisgtr_driver =
   {
      .name = "irisgtr", //Driver name: informational
      .id_table = irisgtr_ids, // device list: irisgtr devices (choose one of the two lines)
      .probe = &irisgtr_dev_init,
      .remove = &irisgtr_remove,
   };

//////////////////////////////////////////////////////////////////////////////
//
// Name         : irisgtr_open
//
// Synopsis     : Driver entry point
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
int irisgtr_open (struct inode *inode, struct file *filp)
   {
   return win_driver_open (iminor(inode), &(filp->private_data));
   }

//////////////////////////////////////////////////////////////////////////////
//
// Name         : irisgtr_release
//
// Synopsis     : Major clean up/ Major close equivalent.
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
int irisgtr_release (struct inode *inode, struct file *filp)
   {
   return win_driver_release (filp, filp->private_data);
   }

//////////////////////////////////////////////////////////////////////////////
//
// Name         : irisgtr_mmap
//
// Synopsis     :
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
int irisgtr_mmap (struct file *filp, struct vm_area_struct *vma)
   {
   vma->vm_flags |= VM_RESERVED | VM_IO;
   if (io_remap_pfn_range
       (vma, vma->vm_start, vma->vm_pgoff, vma->vm_end - vma->vm_start, vma->vm_page_prot))
      return -EAGAIN;

   return 0;
   }


//////////////////////////////////////////////////////////////////////////////
//
//  The following functions fill in the required structures for the /proc
//  entries.
//
//////////////////////////////////////////////////////////////////////////////
static int irisgtr_boardtype_show(struct seq_file *sfile, void *iterator)
   {
   const int BufferSize = 512;
   char *Buffer = MDrv_memalloc(BufferSize);
   win_get_board_type((unsigned long)sfile->private, Buffer, BufferSize);
   seq_printf(sfile, "%s", Buffer);
   MDrv_memfree(Buffer);
   return 0;
   }

static int irisgtr_boardtype_open(struct inode *inode, struct file *file)
   {
   return single_open(file, irisgtr_boardtype_show, PDE_DATA(inode));
   }

static const struct file_operations irisgtr_boardtype_ops = {
   .owner   = THIS_MODULE,
   .open    = irisgtr_boardtype_open,
   .read    = seq_read,
   .llseek  = seq_lseek,
   .release = single_release
};

static int irisgtr_dbglevel_show(struct seq_file *sfile, void *iterator)
   {
   seq_printf(sfile, "0x%x\n", MDrv_DbgLevel);
   return 0;
   }

static int irisgtr_dbglevel_open(struct inode *inode, struct file *file)
   {
   return single_open(file, irisgtr_dbglevel_show, PDE_DATA(inode));
   }

static ssize_t irisgtr_dbglevel_write_proc(struct file *file, const char __user *buffer,
                                       size_t count, loff_t *data)
   {
   char *localBuf = NULL, *tmp = NULL;
   int newcount = count;
   int val = 0;
   int ret = 0;

   localBuf = kmalloc(count+1, GFP_KERNEL);

   if(localBuf == NULL)
      return -ENOMEM;

   tmp = localBuf;

   while((ret = copy_from_user(tmp, buffer, newcount)))
      {
      tmp += newcount - ret;
      buffer += newcount - ret;
      newcount -= newcount - ret;
      }

   localBuf[count] = '\0';

   sscanf(localBuf, "%x\n", &val);

   MDrv_DbgLevel = val;

   kfree(localBuf);
   return count;
   }

static const struct file_operations irisgtr_dbglevel_ops = {
   .owner   = THIS_MODULE,
   .open    = irisgtr_dbglevel_open,
   .read    = seq_read,
   .write   = irisgtr_dbglevel_write_proc,
   .llseek  = seq_lseek,
   .release = single_release
};

static int irisgtr_nbBoards_show(struct seq_file *sfile, void *iterator)
   {
   seq_printf(sfile,"%lu\n", nbOfBoards);
   return 0;
   }

static int irisgtr_nbBoards_open(struct inode *inode, struct file *file)
   {
   return single_open(file, irisgtr_nbBoards_show, PDE_DATA(inode));
   }

static const struct file_operations irisgtr_nbBoards_ops = {
   .owner   = THIS_MODULE,
   .open    = irisgtr_nbBoards_open,
   .read    = seq_read,
   .llseek  = seq_lseek,
   .release = single_release
};

static int irisgtr_builddate_show(struct seq_file *sfile, void *iterator)
   {
   seq_printf(sfile,"Module built on %s at %s\n",  __DATE__, __TIME__);
   return 0;
   }

static int irisgtr_builddate_open(struct inode *inode, struct file *file)
   {
   return single_open(file, irisgtr_builddate_show, PDE_DATA(inode));
   }

static const struct file_operations irisgtr_builddate_ops = {
   .owner   = THIS_MODULE,
   .open    = irisgtr_builddate_open,
   .read    = seq_read,
   .llseek  = seq_lseek,
   .release = single_release
};

static int irisgtr_logbuffer_show(struct seq_file *sfile, void *iterator)
   {
   seq_printf(sfile,"0x%lx\n%lu\n",  lxGetLogBufferAddr(), lxGetLogBufferSize());
   return 0;
   }

static int irisgtr_logbuffer_open(struct inode *inode, struct file *file)
   {
   return single_open(file, irisgtr_logbuffer_show, PDE_DATA(inode));
   }

static const struct file_operations irisgtr_logbuffer_ops = {
   .owner   = THIS_MODULE,
   .open    = irisgtr_logbuffer_open,
   .read    = seq_read,
   .llseek  = seq_lseek,
   .release = single_release
};

static int irisgtr_devicepresent_show(struct seq_file *sfile, void *iterator)
   {
   const int DevicePresent = 1;
   seq_printf(sfile, "0x%X\n", DevicePresent);
   return 0;
   }

static int irisgtr_devicepresent_open(struct inode *inode, struct file *file)
   {
   return single_open(file, irisgtr_devicepresent_show, PDE_DATA(inode));
   }

static const struct file_operations irisgtr_devicepresent_ops = {
   .owner   = THIS_MODULE,
   .open    = irisgtr_devicepresent_open,
   .read    = seq_read,
   .llseek  = seq_lseek,
   .release = single_release
};

static int irisgtr_nbofpath_show(struct seq_file *sfile, void *iterator)
   {
   const int BufferSize = 512;
   char *Buffer = MDrv_memalloc(BufferSize);
   win_get_nbofpath((unsigned long)sfile->private, Buffer, BufferSize);
   seq_printf(sfile, "%s", Buffer);
   MDrv_memfree(Buffer);
   return 0;
   }

static int irisgtr_nbofpath_open(struct inode *inode, struct file *file)
   {
   return single_open(file, irisgtr_nbofpath_show, PDE_DATA(inode));
   }

static const struct file_operations irisgtr_nbofpath_ops = {
   .owner   = THIS_MODULE,
   .open    = irisgtr_nbofpath_open,
   .read    = seq_read,
   .llseek  = seq_lseek,
   .release = single_release
};

//////////////////////////////////////////////////////////////////////////////
//
//  The following functions enable and disable the /proc entries.
//
//////////////////////////////////////////////////////////////////////////////
void irisgtr_create_board_proc_entries(unsigned long boardNb)
   {
   char name[100];
   struct proc_dir_entry *board_entry;
   snprintf(name, 99, "dev%lu", boardNb );
   board_entry = proc_mkdir(name, irisgtr_volatile_dir);
   proc_create_data("boardtype", 0444, board_entry, &irisgtr_boardtype_ops, (void *) boardNb);
   proc_create_data("devicepresent", 0444, board_entry, &irisgtr_devicepresent_ops, (void *) boardNb);
   proc_create_data("nbofacquisitionpath", 0444, board_entry, &irisgtr_nbofpath_ops, (void *) boardNb);
   }

void irisgtr_init_proc(void)
   {
   struct proc_dir_entry *entry;
   int i = 0;

   irisgtr_proc_dir = proc_mkdir("driver/mil/irisgtr", NULL);
   irisgtr_volatile_dir = proc_mkdir("volatile", irisgtr_proc_dir);

   entry = proc_create_data("dbg_level", 0666, irisgtr_proc_dir, &irisgtr_dbglevel_ops, NULL);
   entry = proc_create_data("nb_boards", 0444, irisgtr_proc_dir, &irisgtr_nbBoards_ops, NULL);
   entry = proc_create_data("build_date", 0444, irisgtr_proc_dir, &irisgtr_builddate_ops, NULL);
   entry = proc_create_data("log_buffer", 0444, irisgtr_proc_dir, &irisgtr_logbuffer_ops, NULL);

   for( i = 0; i < nbOfBoards; i++)
      {
      irisgtr_create_board_proc_entries(i);
      }
   }

void irisgtr_remove_proc(void)
   {
   char entry[100];
   unsigned long i = 0;

   remove_proc_entry("nb_boards", irisgtr_proc_dir);
   remove_proc_entry("dbg_level", irisgtr_proc_dir);
   remove_proc_entry("build_date", irisgtr_proc_dir);
   remove_proc_entry("log_buffer", irisgtr_proc_dir);

   for(i = 0; i < nbOfBoards; i++)
      {
      snprintf(entry, 99, "dev%lu/boardtype", i);
      remove_proc_entry(entry, irisgtr_volatile_dir);
      snprintf(entry, 99, "dev%lu/devicepresent", i);
      remove_proc_entry(entry, irisgtr_volatile_dir);
      snprintf(entry, 99, "dev%lu/nbofacquisitionpath", i);
      remove_proc_entry(entry, irisgtr_volatile_dir);
      snprintf(entry, 99, "dev%lu", i);
      remove_proc_entry(entry, irisgtr_volatile_dir);
      }
   remove_proc_entry("volatile", irisgtr_proc_dir);
   remove_proc_entry("driver/mil/irisgtr", NULL);
   }

//////////////////////////////////////////////////////////////////////////////
//
// Name         : irisgtr_unlocked_ioctl
//
// Synopsis     : DrvDispatch
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
long irisgtr_unlocked_ioctl (struct file *filp, unsigned int cmd, unsigned long arg)
   {
   return win_driver_ioctl (cmd, arg, filp->private_data);
   }

//////////////////////////////////////////////////////////////////////////////
//
// Name         : irisgtr_dev_init
//
// Synopsis     : Detect irisgtr board
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
int irisgtr_dev_init (struct pci_dev *dev, const struct pci_device_id *id)
   {
   u8  myirq  = 0;

   DbgTrace ("In dev_init.\n");
   DbgTrace ("   vendor:      0x%X\n", (unsigned int) dev->vendor);
   DbgTrace ("   device:      0x%X\n", (unsigned int) dev->device);
   DbgTrace ("   subvendor:   0x%X\n", (unsigned int) dev->subsystem_vendor);
   DbgTrace ("   subdevice:   0x%X\n", (unsigned int) dev->subsystem_device);
   //DbgTrace ("   class:       0x%X\n", (unsigned int) dev->class_tmp);
   DbgTrace ("   bus:         0x%X\n", (unsigned int) dev->bus->number);
#ifdef CONFIG_PCI_NAMES
   DbgTrace ("   name:        %s\n",   dev->pretty_name);
#else
   DbgTrace ("   name:        unknown (no pci names kernel support)");
#endif
   DbgTrace ("   driver_data: 0x%X\n", (unsigned int) id->driver_data);
   DbgTrace ("   Bus number:  %u\n",   (unsigned char)dev->bus->number);

   myirq = read_config_byte (dev, PCI_INTERRUPT_LINE);

   DbgTrace ("   The interrupt: %d\n", (unsigned int)myirq);

   if (dev->bus)
      {
      if (pci_enable_device (dev))
         {
         DbgTrace ("cannot enable device\n");
         return -ENODEV;
         }

      if (pci_enable_msi (dev) != 0)
         {
         DbgTrace ("cannot enable MSI interrupts\n");
         return -ENODEV;
         }

      MDrv_pcidev_t md_pci_dev;
      ospci2mdpci(dev, &md_pci_dev);

      if (win_driver_device_init (&md_pci_dev, NULL) != 0)
         return -ENODEV;

      pci_set_master (dev);

      DbgTrace ("Device succesfully registered.\n");
      return 0;
      }
   else
      {
      DbgTrace ("Proposed device is not on a bus. Rejecting.\n");
      return -ENODEV;
      }
   }

void irisgtr_remove (struct pci_dev *dev)
{
   pci_disable_msi (dev);
}
//////////////////////////////////////////////////////////////////////////////
//
// Name         : irisgtr_cleanup
//
// Synopsis     : undo whatever irisgtr_init did
//
// Parameters   : none
//
// Return Value : none
//
//////////////////////////////////////////////////////////////////////////////
void irisgtr_cleanup (void)
   {
   DbgTrace ("Cleaning the irisgtr module\n");
   irisgtr_remove_proc ();

   win_driver_cleanup ();
   DbgTrace ("Unregistering device\n");
   // Unregister device driver
   unregister_chrdev_region (irisgtr_major, IRISGTR_MAX_CARDS);
   pci_unregister_driver (&irisgtr_driver);

   DbgTrace ("Unloading driver object\n");
   win_driver_unload ();

   DbgTrace ("Matrox irisgtr driver is unloaded\n");
   }


//////////////////////////////////////////////////////////////////////////////
//
// Name         : irisgtr_init
//
// Synopsis     : Initialize irisgtr linux driver
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
int irisgtr_init (void)
   {
   int err = 0;

   DbgTrace ("%s - Matrox irisgtr driver is loading...\n", MDrv_DriverName);

   if (alloc_chrdev_region (&irisgtr_major, 0, IRISGTR_MAX_CARDS, "Matrox irisgtr") < 0)
      {
      DbgPrint ("%s - Can't get major number %d\n", MDrv_DriverName, MAJOR(irisgtr_major));
      irisgtr_cleanup ();
      return -ENOMEM;
      }

   if (win_driver_init () != 0)
      return -ENOMEM;

   err = pci_register_driver (&irisgtr_driver);
   if (err < 0)
      {
      irisgtr_cleanup ();
      return -ENODEV;
      }

   irisgtr_init_proc();
   printk ("irisgtr - Module built on %s at %s loaded successfully.\n", __DATE__, __TIME__);
   return err;
   }

module_init(irisgtr_init);
module_exit(irisgtr_cleanup);
