/***************************************************************************
 *
 * Rev          :  $Revision:   1.0  $
 * Content      :  Entry (and exit) points for the Linux Driver
 * Compile flag :
 *
 * COPYRIGHT (c) 2003  Matrox Electronic Systems Ltd.
 * All Rights Reserved
 *
 ****************************************************************************/
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/kallsyms.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/uaccess.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include "linuxentry.h"
#include "linux-interface.h"

#include <linux/types.h>
#include <linux/acpi.h>
#include <acpi/acpi_bus.h>
#include <acpi/acpi_drivers.h>

const char *MDrv_DriverName = "MTXAUX";
int MDrv_DbgLevel = DBG_CRIT|DBG_ERROR;
int mtxaux_major = 0;
int indio_major = 0;
typedef enum {
    initial,
    mtxaux_acpi,
    mtxaux_pci,
    }mtxaux_driver_mode;
mtxaux_driver_mode drvmode = initial;

extern unsigned long nbOfBoards;
unsigned long nbOfMtxAux;
unsigned long nbOfIndio;
static struct proc_dir_entry *mtxaux_proc_dir;
static struct proc_dir_entry *mtxaux_volatile_dir;
static struct proc_dir_entry *indio_proc_dir;
static struct proc_dir_entry *indio_volatile_dir;

MODULE_LICENSE("Proprietary");
MODULE_DESCRIPTION("MIL Mtx Auxiliary Driver");
MODULE_AUTHOR("Matrox Electronic Systems Ltd.");

static long mtxaux_unlocked_ioctl (struct file *filp, unsigned int cmd, unsigned long arg);
static int mtxaux_open (struct inode *inode, struct file *filp);
static int mtxaux_mmap (struct file *filp, struct vm_area_struct *vma);
static int mtxaux_release (struct inode *inode, struct file *filp);
static int mtxaux_pci_probe (struct pci_dev *dev, const struct pci_device_id *id);
static void mtxaux_cleanup (void);
static int mtxaux_init (void);

struct file_operations mtxaux_fops =
    {
        .unlocked_ioctl = mtxaux_unlocked_ioctl,
        .open = mtxaux_open,
        .release = mtxaux_release,
        .mmap = mtxaux_mmap
    };

//Structure defining all the IDs that belong to nexis2
struct pci_device_id mtxaux_pci_ids[] =
    {
    { MATROX_VENDID, FOURSIGHT_DEVICE_ID, MATROX_VENDID, FOURSIGHT_SUBSYS_ID, 0, 0, 0},
    { MATROX_VENDID, FOURSIGHT_DEVICE_ID, MATROX_VENDID, INDIOAUX_SUBSYS_ID, 0, 0, 0},
    { MATROX_VENDID, FOURSIGHT_DEVICE_ID, MATROX_VENDID, IRISGTR_SUBSYS_ID, 0, 0, 0},
    {0,}
    };

//Structure defining the nexis2 driver for hotplug.
struct pci_driver mtxaux_pci_driver =
    {
        .name = "MtxAux", //Dirver name: informational
        .id_table = mtxaux_pci_ids, // device list: mtxaux devices (choose one of the two lines)
        .probe = &mtxaux_pci_probe
    };


static const struct acpi_device_id mtxaux_acpi_id [] =
    {
    { "MTX5E10", 0 },
    { "MTX5E11", 0 },
    { "", 0 },
    };
MODULE_DEVICE_TABLE(acpi, mtxaux_acpi_id);

static int mtxaux_acpi_add(struct acpi_device *device);
static int mtxaux_acpi_remove(struct acpi_device *device);

static struct acpi_driver mtxaux_acpi_driver =
    {
        .name = "MtxAux",
        .class = "Matrox Imaging",
                .ids = mtxaux_acpi_id,
                .ops =
            {
                .add = mtxaux_acpi_add,
                .remove = mtxaux_acpi_remove,
            },
    };

////////////////////////////////////////////////////////////////////////////
//
// Name         : Mtxaux_open
//
// Synopsis     : Driver entry point
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
////////////////////////////////////////////////////////////////////////////
static int mtxaux_open (struct inode *inode, struct file *filp)
    {
    return win_driver_open (iminor(inode), &(filp->private_data));
    }

//////////////////////////////////////////////////////////////////////////////
//
// Name         : mtxaux_release
//
// Synopsis     : Major clean up/ Major close equivalent.
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
static int mtxaux_release (struct inode *inode, struct file *filp)
    {
    return win_driver_release (filp, filp->private_data);
    }


//////////////////////////////////////////////////////////////////////////////
//
// Name         : mtxaux_mmap
//
// Synopsis     :
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
static int mtxaux_mmap (struct file *filp, struct vm_area_struct *vma)
    {
    vma->vm_flags |= VM_RESERVED | VM_IO;
    if (io_remap_pfn_range
            (vma, vma->vm_start, vma->vm_pgoff, vma->vm_end - vma->vm_start, vma->vm_page_prot))
        return -EAGAIN;

    return 0;
    }

//////////////////////////////////////////////////////////////////////////////
//
// Name         : mtxaux_unlocked_ioctl
//
// Synopsis     : DrvDispatch
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
static long mtxaux_unlocked_ioctl (struct file *filp, unsigned int cmd, unsigned long arg)
    {
    return win_driver_ioctl (cmd, arg, filp->private_data);
    }


//////////////////////////////////////////////////////////////////////////////
//
//  The following functions fill in the required structures for the /proc
//  entries.
//
//////////////////////////////////////////////////////////////////////////////
static int mtxaux_boardtype_show(struct seq_file *sfile, void *iterator)
    {
    const int BufferSize = 512;
    char *Buffer = MDrv_memalloc(BufferSize);
    win_get_board_type((unsigned long)sfile->private, Buffer, BufferSize);
    seq_printf(sfile, "%s", Buffer);
    MDrv_memfree(Buffer);
    return 0;
    }

static int mtxaux_boardtype_open(struct inode *inode, struct file *file)
    {
    return single_open(file, mtxaux_boardtype_show, PDE_DATA(inode));
    }

static const struct file_operations mtxaux_boardtype_ops = {
        .owner   = THIS_MODULE,
        .open    = mtxaux_boardtype_open,
        .read    = seq_read,
        .llseek  = seq_lseek,
        .release = single_release
    };

static int mtxaux_dbglevel_show(struct seq_file *sfile, void *iterator)
    {
    seq_printf(sfile, "0x%x\n", MDrv_DbgLevel);
    return 0;
    }

static int mtxaux_dbglevel_open(struct inode *inode, struct file *file)
    {
    return single_open(file, mtxaux_dbglevel_show, PDE_DATA(inode));
    }

static ssize_t mtxaux_dbglevel_write_proc(struct file *file, const char __user *buffer,
                                          size_t count, loff_t *data)
    {
    char *localBuf = NULL, *tmp = NULL;
    int newcount = count;
    int val = 0;
    int ret = 0;

    localBuf = kmalloc(count+1, GFP_KERNEL);

    if(localBuf == NULL)
        return -ENOMEM;

    tmp = localBuf;

    while((ret = copy_from_user(tmp, buffer, newcount)))
        {
        tmp += newcount - ret;
        buffer += newcount - ret;
        newcount -= newcount - ret;
        }

    localBuf[count] = '\0';

    sscanf(localBuf, "%x\n", &val);

    MDrv_DbgLevel = val;

    kfree(localBuf);
    return count;
    }

static const struct file_operations mtxaux_dbglevel_ops = {
        .owner   = THIS_MODULE,
        .open    = mtxaux_dbglevel_open,
        .read    = seq_read,
        .write   = mtxaux_dbglevel_write_proc,
        .llseek  = seq_lseek,
        .release = single_release
    };

static int mtxaux_nbBoards_show(struct seq_file *sfile, void *iterator)
    {
    seq_printf(sfile,"%lu\n", nbOfMtxAux);
    return 0;
    }

static int mtxaux_nbBoards_open(struct inode *inode, struct file *file)
    {
    return single_open(file, mtxaux_nbBoards_show, PDE_DATA(inode));
    }

static const struct file_operations mtxaux_nbBoards_ops = {
        .owner   = THIS_MODULE,
        .open    = mtxaux_nbBoards_open,
        .read    = seq_read,
        .llseek  = seq_lseek,
        .release = single_release
    };

static int indio_nbBoards_show(struct seq_file *sfile, void *iterator)
    {
    seq_printf(sfile,"%lu\n", nbOfIndio);
    return 0;
    }

static int indio_nbBoards_open(struct inode *inode, struct file *file)
    {
    return single_open(file, indio_nbBoards_show, PDE_DATA(inode));
    }
static const struct file_operations indio_nbBoards_ops = {
        .owner   = THIS_MODULE,
        .open    = indio_nbBoards_open,
        .read    = seq_read,
        .llseek  = seq_lseek,
        .release = single_release
    };

static int mtxaux_builddate_show(struct seq_file *sfile, void *iterator)
    {
    seq_printf(sfile,"Module built on %s at %s\n",  __DATE__, __TIME__);
    return 0;
    }

static int mtxaux_builddate_open(struct inode *inode, struct file *file)
    {
    return single_open(file, mtxaux_builddate_show, PDE_DATA(inode));
    }

static const struct file_operations mtxaux_builddate_ops = {
        .owner   = THIS_MODULE,
        .open    = mtxaux_builddate_open,
        .read    = seq_read,
        .llseek  = seq_lseek,
        .release = single_release
    };

static int mtxaux_logbuffer_show(struct seq_file *sfile, void *iterator)
    {
    seq_printf(sfile,"0x%lx\n%lu\n",  lxGetLogBufferAddr(), lxGetLogBufferSize());
    return 0;
    }

static int mtxaux_logbuffer_open(struct inode *inode, struct file *file)
    {
    return single_open(file, mtxaux_logbuffer_show, PDE_DATA(inode));
    }

static const struct file_operations mtxaux_logbuffer_ops = {
        .owner   = THIS_MODULE,
        .open    = mtxaux_logbuffer_open,
        .read    = seq_read,
        .llseek  = seq_lseek,
        .release = single_release
    };

static int mtxaux_devicepresent_show(struct seq_file *sfile, void *iterator)
    {
    const int DevicePresent = 1;
    seq_printf(sfile, "0x%X\n", DevicePresent);
    return 0;
    }

static int mtxaux_devicepresent_open(struct inode *inode, struct file *file)
    {
    return single_open(file, mtxaux_devicepresent_show, PDE_DATA(inode));
    }

static const struct file_operations mtxaux_devicepresent_ops = {
        .owner   = THIS_MODULE,
        .open    = mtxaux_devicepresent_open,
        .read    = seq_read,
        .llseek  = seq_lseek,
        .release = single_release
    };

//////////////////////////////////////////////////////////////////////////////
//
//  The following functions enable and disable the /proc entries.
//
//////////////////////////////////////////////////////////////////////////////
static void mtxaux_create_board_proc_entries(unsigned long boardNb, struct proc_dir_entry * volatile_dir)
    {
    char name[100];
    struct proc_dir_entry *board_entry;
    snprintf(name, 99, "dev%lu", boardNb );
    board_entry = proc_mkdir(name, volatile_dir);
    proc_create_data("boardtype", 0444, board_entry, &mtxaux_boardtype_ops, (void *) boardNb);
    proc_create_data("devicepresent", 0444, board_entry, &mtxaux_devicepresent_ops, (void *) boardNb);
    }

void mtxaux_init_proc(void)
    {
      if(0 == nbOfMtxAux)
      {
        struct proc_dir_entry *entry;
        unsigned long i = 0;
        if (alloc_chrdev_region (&mtxaux_major, 0, MTXAUX_MAX_CARDS, "Matrox Mtxaux") < 0)
            {
            DbgError ("%s - Can't get major number %d\n", MDrv_DriverName, MAJOR(mtxaux_major));
            return -ENOMEM;
            }
        nbOfMtxAux++;
        mtxaux_proc_dir = proc_mkdir("driver/mil/mtxaux", NULL);
        mtxaux_volatile_dir = proc_mkdir("volatile", mtxaux_proc_dir);

        entry = proc_create_data("dbg_level", 0666,  mtxaux_proc_dir, &mtxaux_dbglevel_ops, NULL);
        entry = proc_create_data("nb_boards", 0444,  mtxaux_proc_dir, &mtxaux_nbBoards_ops, NULL);
        entry = proc_create_data("build_date", 0444, mtxaux_proc_dir, &mtxaux_builddate_ops, NULL);
        entry = proc_create_data("log_buffer", 0444, mtxaux_proc_dir, &mtxaux_logbuffer_ops, NULL);

        for( i = 0; i < nbOfMtxAux; i++)
            {
            mtxaux_create_board_proc_entries(i, mtxaux_volatile_dir);
            }
      }
    }

void indio_init_proc(void)
    {
        if(0 == nbOfIndio)
        {
        struct proc_dir_entry *entry;
        unsigned long i = 0;
        if (alloc_chrdev_region (&indio_major, 0, MTXAUX_MAX_CARDS, "Matrox Indio") < 0)
            {
            DbgError ("%s - Can't get major number %d\n", MDrv_DriverName, MAJOR(mtxaux_major));
            return -ENOMEM;
            }
        nbOfIndio++;
        indio_proc_dir = proc_mkdir("driver/mil/indio", NULL);
        indio_volatile_dir = proc_mkdir("volatile", indio_proc_dir);

        entry = proc_create_data("dbg_level", 0666,  indio_proc_dir, &mtxaux_dbglevel_ops, NULL);
        entry = proc_create_data("nb_boards", 0444,  indio_proc_dir, &indio_nbBoards_ops, NULL);
        entry = proc_create_data("build_date", 0444, indio_proc_dir, &mtxaux_builddate_ops, NULL);
        entry = proc_create_data("log_buffer", 0444, indio_proc_dir, &mtxaux_logbuffer_ops, NULL);

        for( i = 0; i < nbOfIndio; i++)
            {
            mtxaux_create_board_proc_entries(i, indio_volatile_dir);
            }
        }
    }


static void mtxaux_remove_proc(void)
    {
    char entry[100];
    unsigned long i = 0;

    if(nbOfMtxAux)
    {
    remove_proc_entry("nb_boards", mtxaux_proc_dir);
    remove_proc_entry("dbg_level", mtxaux_proc_dir);
    remove_proc_entry("build_date", mtxaux_proc_dir);
    remove_proc_entry("log_buffer", mtxaux_proc_dir);

    for(i = 0; i < nbOfMtxAux; i++)
        {
        snprintf(entry, 99, "dev%lu/boardtype", i);
        remove_proc_entry(entry, mtxaux_volatile_dir);
        snprintf(entry, 99, "dev%lu/devicepresent", i);
        remove_proc_entry(entry, mtxaux_volatile_dir);
        snprintf(entry, 99, "dev%lu", i);
        remove_proc_entry(entry, mtxaux_volatile_dir);
        }

    remove_proc_entry("volatile", mtxaux_proc_dir);
    remove_proc_entry("driver/mil/mtxaux", NULL);
    DbgTrace ("Unregistering device\n");
    // Unregister device driver
    unregister_chrdev_region (mtxaux_major, MTXAUX_MAX_CARDS);
    }
    if(nbOfIndio)
    {
    remove_proc_entry("nb_boards",  indio_proc_dir);
    remove_proc_entry("dbg_level",  indio_proc_dir);
    remove_proc_entry("build_date", indio_proc_dir);
    remove_proc_entry("log_buffer", indio_proc_dir);

    for(i = 0; i < nbOfIndio; i++)
        {
        snprintf(entry, 99, "dev%lu/boardtype", i);
        remove_proc_entry(entry, indio_volatile_dir);
        snprintf(entry, 99, "dev%lu/devicepresent", i);
        remove_proc_entry(entry, indio_volatile_dir);
        snprintf(entry, 99, "dev%lu", i);
        remove_proc_entry(entry, indio_volatile_dir);
        }

    remove_proc_entry("volatile", indio_proc_dir);
    remove_proc_entry("driver/mil/indio", NULL);

    DbgTrace ("Unregistering device\n");
    // Unregister device driver
    unregister_chrdev_region (indio_major, MTXAUX_MAX_CARDS);
    }

    }

//////////////////////////////////////////////////////////////////////////////
//
// Name         : mtxaux_pci_probe
//
// Synopsis     : Detect mtxaux board
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
static int mtxaux_pci_probe (struct pci_dev *dev, const struct pci_device_id *id)
    {
    u8 myirq = 0;

    DbgTrace ("In dev_init.\n");
    DbgTrace ("   vendor:      0x%X\n", (unsigned int) dev->vendor);
    DbgTrace ("   device:      0x%X\n", (unsigned int) dev->device);
    DbgTrace ("   subvendor:   0x%X\n", (unsigned int) dev->subsystem_vendor);
    DbgTrace ("   subdevice:   0x%X\n", (unsigned int) dev->subsystem_device);
    //   DbgTrace ("   class:       0x%X\n", (unsigned int) dev->class_tmp);
    DbgTrace ("   bus:         0x%X\n", (unsigned int) dev->bus->number);
#ifdef CONFIG_PCI_NAMES
    DbgTrace ("   name:        %s\n",   dev->pretty_name);
#else
    DbgTrace ("   name:        unknown (no pci names kernel support)");
#endif
    DbgTrace ("   driver_data: 0x%X\n", (unsigned int) id->driver_data);
    DbgTrace ("   Bus number:  %u\n",   (unsigned char) dev->bus->number);

    myirq = read_config_byte (dev, PCI_INTERRUPT_LINE);

    DbgTrace ("   The interrupt: %d\n", (unsigned int)myirq);

    if(dev->subsystem_device == INDIOAUX_SUBSYS_ID)
    {
        indio_init_proc();
    }
    else
    {
        mtxaux_init_proc();
    }
    DbgTrace ("Device succesfully registered.\n");

    //First, make sure it is attached to a bus
    if (dev->bus)
        {
        MDrv_pcidev_t md_pci_dev;
        if (pci_enable_device (dev))
            {
            DbgTrace ("cannot enable device\n");
            return -ENODEV;
            }

        ospci2mdpci(dev, &md_pci_dev);
        if (win_driver_device_init (&md_pci_dev, NULL) != 0)
            return -ENODEV;

        pci_set_master (dev);


        return 0;
        }
    else
        {
        DbgTrace ("Proposed device is not on a bus. Rejecting.\n");
        return -ENODEV;
        }
    }

static acpi_status mtxaux_resources( struct acpi_resource *res, void* data)
    {
    struct mtxaux_acpi_resource_data *lpData = (struct mtxaux_acpi_resource_data *)data;
    switch(res->type)
        {
        case ACPI_RESOURCE_TYPE_FIXED_MEMORY32:
            lpData->mem = res->data.fixed_memory32.address;
            lpData->len = res->data.fixed_memory32.address_length;
            DbgTrace ("   _BAS: 0x%X, _LEN:0x%X\n", lpData->mem, lpData->len);
            break;
        case ACPI_RESOURCE_TYPE_EXTENDED_IRQ:
            lpData->irq = res->data.extended_irq.interrupts[0];
            DbgTrace ("   _IRQ producer_consumer: 0x%X\n", res->data.extended_irq.producer_consumer);
            DbgTrace ("   _IRQ triggering       : 0x%X\n", res->data.extended_irq.triggering);
            DbgTrace ("   _IRQ polarity         : 0x%X\n", res->data.extended_irq.polarity);
            DbgTrace ("   _IRQ sharable         : 0x%X\n", res->data.extended_irq.sharable);
/* Debian 7.8 kernel */            
#if LINUX_VERSION_CODE != KERNEL_VERSION(3,2,65)            
            DbgTrace ("   _IRQ wake_capable     : 0x%X\n", res->data.extended_irq.wake_capable);
#endif            
            DbgTrace ("   _IRQ interrupt_count  : 0x%X\n", res->data.extended_irq.interrupt_count);
            DbgTrace ("   _IRQ interrupts[0]    : 0x%X\n", res->data.extended_irq.interrupts[0]);
            break;
        }
    return AE_OK;
    }

static int mtxaux_acpi_add(struct acpi_device *device)
    {
    acpi_status status;
    struct mtxaux_acpi_resource_data lData;

    status = acpi_walk_resources(device->handle, METHOD_NAME__CRS, mtxaux_resources, &lData);
    if(ACPI_FAILURE(status))
        {
        DbgError ("Failed to get ACPI ressources.\n");
        return -ENODEV;
        }

    mtxaux_init_proc();
    if (win_driver_acpi_device_init (&lData) != 0)
        return -ENODEV;

    return 0;
    }

static int mtxaux_acpi_remove(struct acpi_device *device)
    {
    return 0;
    }

//////////////////////////////////////////////////////////////////////////////
//
// Name         : mtxaux_cleanup
//
// Synopsis     : undo whatever mtxaux_init did
//
// Parameters   : none
//
// Return Value : none
//
//////////////////////////////////////////////////////////////////////////////
static void mtxaux_cleanup (void)
    {
    DbgTrace ("Cleaning the Mtxaux module\n");

    mtxaux_remove_proc();

    win_driver_cleanup ();
    if(drvmode & mtxaux_acpi)
        acpi_bus_unregister_driver(&mtxaux_acpi_driver);
    if(drvmode & mtxaux_pci)
        pci_unregister_driver (&mtxaux_pci_driver);

    DbgTrace ("Unloading driver object\n");
    win_driver_unload ();
    DbgTrace ("Matrox Mtxaux driver is unloaded\n");
    }

//////////////////////////////////////////////////////////////////////////////
//
// Name         : mtxaux_init
//
// Synopsis     : Initialize Mtxaux linux driver
//
// Parameters   : none
//
// Return Value : error code (0 = OK)
//
//////////////////////////////////////////////////////////////////////////////
static int mtxaux_init (void)
    {
    int err = 0;

    DbgTrace ("%s - Matrox Mtxaux driver is loading...\n", MDrv_DriverName);

    if (win_driver_init () != 0)
        return -ENOMEM;

    mtxaux_proc_dir = mtxaux_volatile_dir = mtxaux_proc_dir =  mtxaux_volatile_dir = NULL;
    nbOfMtxAux = nbOfIndio = 0;

    if(0 == acpi_bus_register_driver(&mtxaux_acpi_driver))
        drvmode |= mtxaux_acpi;
    if(0 == pci_register_driver (&mtxaux_pci_driver))
        drvmode |= mtxaux_pci;
    if(drvmode == initial)
        {
        DbgError ("Cannot start: %d\n", drvmode);
        mtxaux_cleanup ();
        return -ENODEV;
        }

    printk ("MTXAUX - Module built on %s at %s loaded successfully.\n", __DATE__, __TIME__);
    return err;
    }

void mtxaux_init_from_nexis(void *arg)
    {
    win_driver_init_from_nexis(arg);
    }
EXPORT_SYMBOL(mtxaux_init_from_nexis);

void mtxaux_init_cmd_list_interface(int instance, void** fn, void**cx)
    {
    init_cmd_list_interface(instance, fn, cx);
    }
EXPORT_SYMBOL(mtxaux_init_cmd_list_interface);

module_init(mtxaux_init);
module_exit(mtxaux_cleanup);
