/***************************************************************************
 *
 * Content      : Linux kernel function stub
 *
 * COPYRIGHT (c) 2004  Matrox Electronic Systems Ltd.
 * All Rights Reserved
 *
 ****************************************************************************/
#ifndef _LINUX_INTERFACE_H
#define _LINUX_INTERFACE_H

#include <stdarg.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifndef VM_RESERVED
#define VM_RESERVED (VM_DONTEXPAND | VM_DONTDUMP)
#endif

#define MDRV_FRDONLY       0x00000001

#define MDRV_USERCONTEXT   0x00000001
#define MDRV_SIRQ          0x00000002
#define MDRV_IRQ           0x00000004
#define MDRV_PCI_ANY_ID    (~0)

typedef unsigned char MDrv_uint8;
typedef unsigned short MDrv_uint16;
typedef unsigned int MDrv_uint32;
typedef unsigned long long MDrv_uint64;

typedef char MDrv_int8;
typedef short MDrv_int16;
typedef int MDrv_int32;
typedef long long MDrv_int64;

typedef struct
   {
   unsigned long sec;
   unsigned long usec;
   } MDrv_time_t;

typedef struct
   {
   unsigned int irq;
   unsigned int dev_fn;
   void *os_dev;
   unsigned short dev_id;
   unsigned char bus_nb;
   unsigned char dev_nb;
   unsigned char func_nb;
   } MDrv_pcidev_t;

typedef struct
   {
   void *os_sem;
   } MDrv_sem_t;

typedef struct
   {
   void *os_mutex;
   } MDrv_mutex_t;

typedef struct
   {
   void *os_spin;
   int in_isr;
   } MDrv_spin_t;

typedef struct
   {
   void *os_tasklet;
   } MDrv_tasklet_t;

typedef struct
   {
   void *os_work;
   } MDrv_work_t;

typedef struct
   {
   void *os_compl;
   int signaled;
   } MDrv_compl_t;

typedef struct
   {
   void *os_uart;
   } MDrv_uart_t;
   
/* Debian 7.8 kernel */
#define KERNEL_VERSION_MIL(a,b,c) (((a) << 16) + ((b) << 8) + (c))
#if ( LINUX_VERSION_CODE == KERNEL_VERSION_MIL(3,2,65) )
#define PDE_DATA(A) 0
#endif
  
extern int MDrv_DbgLevel;
extern const char *MDrv_DriverName;
  
#define DBG_CRIT      0x00000001
#define DBG_ERROR     0x00000002
#define DBG_TRACE     0x00000004
#define DBG_PRINT     0x00000008
#define DBG_TMP_TRACE 0x00000010

#define DbgCritical( fmt, args... ) if (MDrv_DbgLevel & DBG_CRIT) MDrv_print ("%s:%s - " fmt, MDrv_DriverName, __func__, ##args )
#define DbgError( fmt, args... ) if (MDrv_DbgLevel & DBG_ERROR) MDrv_print ("%s:%s - " fmt, MDrv_DriverName, __func__, ##args )
#define DbgTrace( fmt, args... ) if (MDrv_DbgLevel & DBG_TRACE) MDrv_print ("%s:%s - " fmt, MDrv_DriverName, __func__, ##args )
#define DbgTmpTrace( fmt, args... ) if (MDrv_DbgLevel & DBG_TMP_TRACE) MDrv_print ("%s:%s - " fmt, MDrv_DriverName, __func__, ##args )
#define DbgPrint( a, args... ) if (MDrv_DbgLevel & DBG_PRINT) MDrv_print ( a, ##args )
#define DbgImp(a, args...)
#define DbgImpMOK(a, args...)
#define DbgAssert( a, fmt, args... ) if(!(a)){MDrv_print ("%s - DEBUG ASSERT %s - " fmt "\n", MDrv_DriverName, #a, ##args); /*ASSERT_BREAKPOINT*/}
#define assert(args...)
#define RETAILMSG(expr, args ) if ((MDrv_DbgLevel & DBG_TRACE) && (expr)) MDrv_print (args)

int MDrv_inatomic(void);
int MDrv_irqsdisabled(void);

void *MDrv_memcpy (void *dest, const void *src, int n);
int   MDrv_memcpy_s (void *dest, unsigned int size, const void *src, unsigned int count);
void *MDrv_memmove (void *dest, const void *src, int n);
int   MDrv_memmove_s (void *dest, unsigned int size, const void *src, unsigned int count);
void *MDrv_memset (void *s, int c, int n);
void *MDrv_memset16 (void *dest, int patern, int count);
void *MDrv_memset32 (void *dest, int patern, int count);
int MDrv_memcmp(const void *s1, const void *s2, int n);

#define MOs_ANSIstrlen MOs_strlen
int MOs_strlen (const char *s);
int MOs_strnlen (const char *s, unsigned int size);

#define MOs_ANSIstrcpy_s MOs_strcpy_s
char *MOs_strcpy (char *dest, const char *src);
int   MOs_strcpy_s (char *dest, unsigned int size, const char *src);
char *MOs_strncpy (char *dest, const char *src, unsigned int size);
int   MOs_strncpy_s (char *dest, unsigned int size, const char *src, unsigned int count);

#define MOs_ANSIstrcat_s MOs_strcat_s
char *MOs_strcat (char *dest, const char *src);
int   MOs_strcat_s (char *dest, unsigned int size, const char *src);
char *MOs_strncat (char *dest, const char *src, unsigned int size);
int   MOs_strncat_s (char *dest, unsigned int size, const char *src, unsigned int count);

int MOs_strcmp (const char *s1, const char *s2);
int MOs_stricmp(const char *s1, const char *s2);
#define MOs_ANSIstrncmp MOs_strncmp
int MOs_strncmp (const char *s1, const char *s2, int n);

int MOs_strtolower (unsigned char *Buffer);

int MDrv_print (const char *format, ...);
#define MOs_ANSIsprintf_s(A, B, ...) MOs_sprintf(A, __VA_ARGS__)

#define MOs_sprintf_s(A, B, ...) MOs_sprintf(A, __VA_ARGS__)
int MOs_sprintf (char *str, const char *format, ...);
#define MOs_ANSIvsprintf_s MOs_vsprintf_s
int MOs_vsprintf_s (char *str, int size, const char *format, va_list list);
int MDrv_snprintf (char *str, int size, const char *format, ...);
int MDrv_sscanf (const char *str, const char *format, ...);

void *MDrv_memalloc (int size);
void MDrv_memfree (const void *p);
void *MDrv_vmemalloc (int size);
void MDrv_vmemfree (void *p);
unsigned long MDrv_getpage(int size);
void MDrv_freepage(unsigned long addr);

unsigned long MDrv_allocate_stack (void **StackPtr, unsigned long StackSize);
void MDrv_free_stack (void * StackPtr);

void *MDrv_fopen (const char *filename, int attr);
unsigned long MDrv_fsize (void *file);
int MDrv_fread (void *file, unsigned long fpos, char *buf, unsigned long bufsize);
int MDrv_fclose (void *file);

int MDrv_read_dword_registry(const char* regEntry, MDrv_uint32 *Data);

unsigned long MDrv_getstate (void);
void MDrv_raisestate (unsigned long newstate, unsigned long *oldstate);
void MDrv_lowerstate (unsigned long newstate);

void *MDrv_map (unsigned long addr, unsigned long size, int NoErrorPrint);
void *MDrv_map_cache (unsigned long addr, unsigned long size, int NoErrorPrint);
void *MDrv_map_wc (unsigned long addr, unsigned long size, int NoErrorPrint);
void MDrv_unmap (void *p);
void *MDrv_requestmem (unsigned long addr, unsigned long size, const char *ownername);
void *MDrv_requestio (unsigned long addr, unsigned long size, const char *ownername);
void MDrv_releasemem (unsigned long addr, unsigned long size);
void MDrv_releaseio (unsigned long addr, unsigned long size);
void MDrv_releaseresource (void *Resource);
unsigned long MDrv_physaddr (void *virtaddr);

void MDrv_gettime (MDrv_time_t *curtime);

int MDrv_getcurrentprocid (void);
int MDrv_getcurrentthreadid (void);
void MDrv_wakeprocess (void *task);

MDrv_uint64 MDrv_getcounter64 (void);
int MDrv_getcounterfreq (void);
MDrv_uint64 MDrv_getjiffies64 (void);
int MDrv_getjiffiesfreq (void);

void ospci2mdpci (void *os, MDrv_pcidev_t *md);
int MDrv_getpcidevicebyid (unsigned short DeviceId, unsigned short VendorId, unsigned short Index, MDrv_pcidev_t *pcidev);
int MDrv_getnextpcidevicebyid (unsigned short DeviceId, unsigned short VendorId, unsigned short Index, MDrv_pcidev_t *pcidev);
int MDrv_getpcidevicebynb (unsigned char BusNb, unsigned char DeviceNb, unsigned char FunctNb, MDrv_pcidev_t *pcidev);
int MDrv_pci_read_config_byte (MDrv_pcidev_t *pci_device, int where, MDrv_uint8 *Val);
int MDrv_pci_read_config_word (MDrv_pcidev_t *pci_device, int where, MDrv_uint16 *Val);
int MDrv_pci_read_config_dword (MDrv_pcidev_t *pci_device, int where, MDrv_uint32 *Val);
int MDrv_pci_write_config_byte (MDrv_pcidev_t *pci_device, int where, MDrv_uint8 Val);
int MDrv_pci_write_config_word (MDrv_pcidev_t *pci_device, int where, MDrv_uint16 Val);
int MDrv_pci_write_config_dword (MDrv_pcidev_t *pci_device, int where, MDrv_uint32 Val);
void MDrv_pci_dereference(MDrv_pcidev_t *pci_device);
unsigned long MDrv_pciresourcestart (MDrv_pcidev_t *dev, int bar);
unsigned long MDrv_pciresourcelen (MDrv_pcidev_t *dev, int bar);
int MDrv_pciresourceismem(MDrv_pcidev_t*dev, int bar);
int MDrv_pciresourceisio(MDrv_pcidev_t*dev, int bar);

int MDrv_delaythread (MDrv_uint64 interval, int alertable);
void MDrv_msleep(unsigned int milisecond);
void MDrv_usleep(unsigned long microsecond);
void MDrv_nsleep(unsigned long nanosecond);

void MDrv_seminit (MDrv_sem_t *sem, int count);
void MDrv_semdelete (MDrv_sem_t *sem);
void MDrv_semmutexinit (MDrv_sem_t *sem);
int MDrv_semdown (MDrv_sem_t *sem);
int MDrv_semdowninterruptible (MDrv_sem_t *sem);
int MDrv_semtry (MDrv_sem_t *sem);
int MDrv_semdowntimeout (MDrv_sem_t *sem, unsigned long timeout);
void MDrv_semup (MDrv_sem_t *sem);

void MDrv_mutexinit(MDrv_mutex_t *mutex);
void MDrv_mutexdelete(MDrv_mutex_t *mutex);
void MDrv_mutexlock(MDrv_mutex_t *mutex);
int MDrv_mutexlockinterruptible(MDrv_mutex_t *mutex);
int MDrv_mutextrylock(MDrv_mutex_t *mutex);
void MDrv_mutexunlock(MDrv_mutex_t *mutex);

void MDrv_spinlockinit (MDrv_spin_t *lock);
void MDrv_spinlockdelete (MDrv_spin_t *lock);
void MDrv_spinlock (MDrv_spin_t *lock);
void MDrv_spinunlock (MDrv_spin_t *lock);
void MDrv_spinlock_bh (MDrv_spin_t *lock);
void MDrv_spinunlock_bh (MDrv_spin_t *lock);
void MDrv_spinlock_irqsave (MDrv_spin_t *lock, unsigned long *flags);
void MDrv_spinunlock_irqrestore (MDrv_spin_t *lock, unsigned long *flags);

void MDrv_disablepreempt (void);
void MDrv_enablepreempt (void);

int MDrv_requestirq (unsigned int irq, int (*handler)(int, void *), const char *devname, void *dev_id);
void MDrv_freeirq (unsigned int irq, void *dev_id);
void MDrv_synchronizeirq (unsigned int irq);

void MDrv_inittasklet (MDrv_tasklet_t *task, void (*func)(unsigned long), unsigned long data);
void MDrv_deletetasklet (MDrv_tasklet_t *task);
int MDrv_scheduletasklet (MDrv_tasklet_t *task);
void MDrv_canceltasklet(MDrv_tasklet_t *task);
void MDrv_init_work (MDrv_work_t *task, void (*func)(unsigned long), unsigned long data);
void MDrv_delete_work(MDrv_work_t *task);
int MDrv_schedule_work (MDrv_work_t *task);
int MDrv_cancel_work(MDrv_work_t *task);
void MDrv_flush_work(void);

void *MDrv_startthread (int (*fn)(void *), void * arg);
int MDrv_stopthread (void *thread);
int MDrv_threadshouldstop (void);

unsigned int MDrv_readl (const void *addr);

MDrv_uint8 read_config_byte (void *dev, int address);
MDrv_uint16 read_config_word (void *dev, int address);
MDrv_uint32 read_config_dword (void *dev, int address);

int MDrv_copytouser (unsigned char *user, unsigned char *kernel, int size);
int MDrv_copyfromuser (unsigned char *kernel, unsigned char *user, int size);
void MDrv_setinterruptible (void);
void MDrv_setrunning (void);
void MDrv_wmb(void);
void MDrv_schedule (void);
void *MDrv_getcurrenttask (void);

void *MDrv_adddevice (long major, long offset, void *fops, const char *cdev_name, const char *devfs_name);
void MDrv_removedevice (void *dev_handle, long offset, const char *devfs_name);

void MDrv_flushsignals (void);

void MDrv_dumpstack (void);
void MDrv_printsymbol (unsigned long addr);

int MDrv_mtrr_add (void* addr, long size, int flags);
int MDrv_mtrr_del (int reg, void* addr, long size);

void MDrv_initcompletion (MDrv_compl_t *comp);
void MDrv_reinitcompletion (MDrv_compl_t *comp);
void MDrv_deletecompletion (MDrv_compl_t *comp);
void MDrv_waitcompletion (MDrv_compl_t *comp);
int MDrv_waitcompletion_interruptible (MDrv_compl_t *comp);
unsigned long MDrv_waitcompletion_timeout (MDrv_compl_t *comp, unsigned  long timeout);
unsigned long MDrv_waitcompletion_interruptible_timeout (MDrv_compl_t *comp, unsigned  long timeout);
void MDrv_signalcompletion (MDrv_compl_t *comp);

int MDrv_alloc_uart_port (MDrv_uart_t *uart);
void MDrv_free_uart_port (MDrv_uart_t *uart);

unsigned char MDrv_inb(unsigned long port);
void MDrv_outb(unsigned char value, unsigned long port);
unsigned short MDrv_inw(unsigned long port);
void MDrv_outw(unsigned short value, unsigned long port);
unsigned int MDrv_inl(unsigned long port);
void MDrv_outl(unsigned int value, unsigned long port);
void MDrv_atomic_test(void);

unsigned long long MDrv_read_msr(unsigned int msr);
unsigned long MDrv_popcountl(unsigned long value);

int MDrv_is_rt_kernel(void);

typedef int (*SerialISRCallback) (MDrv_pcidev_t *arg);
typedef SerialISRCallback (*ConfigureSerialCallback) (MDrv_pcidev_t *dev);
int win_driver_open (int SystemIndex, void **data);
int win_driver_release (void *fileobject, void *data);
int win_driver_ioctl (unsigned int cmd, unsigned long arg, void *filp_private_data);
int win_driver_device_init (MDrv_pcidev_t *dev, ConfigureSerialCallback ConfigureSerial);
int win_driver_cleanup (void);
int win_driver_unload (void);
int win_driver_init (void);
int win_get_board_type(unsigned long nb, char *buf, int count);
unsigned long lxGetLogBufferAddr(void);
unsigned long lxGetLogBufferSize(void);

#ifdef __cplusplus
}
#endif

#endif // _LINUX_INTERFACE_H



