/***************************************************************************
 *
 * Filename     :  linuxentry.cpp
 * Content      :  Entry (and exit) points for the Linux MTXSERVMANAGER Driver
 *
 * COPYRIGHT (c) 2006  Matrox Electronic Systems Ltd.
 * All Rights Reserved
 *
 ****************************************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/uaccess.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/unistd.h>
#include <asm/bitops.h>
//#include <asm/semaphore.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/interrupt.h>
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/ioctl.h>
#include <asm/msr.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <linux/list.h>
#include <linux/seq_file.h>

#include "linuxentry.h"
#include "linux-interface.h"

#define MAX_DEVICE_OBJECT 1
typedef struct _MSM_DRIVER_INSTANCE
{
   struct cdev       cdev;
   void *msm_device_object;
} MSM_DRIVER_INSTANCE;

const char *MDrv_DriverName = "MTXSERVMANAGER";
MODULE_LICENSE("Proprietary");
MODULE_DESCRIPTION("Matrox Imaging service manager Module");
MODULE_AUTHOR("Matrox Electronic Systems Ltd.");

int MDrv_DbgLevel = DBG_CRIT|DBG_ERROR;
static dev_t   DevId = 0;
static MSM_DRIVER_INSTANCE*  MSMGeneralDevice = NULL;

static unsigned int GlobalASPMDone = 0xFFFFFFFF;
static unsigned int GlobalASPMFailed = 0xFFFFFFFF;
static unsigned int GlobalASPMSeen = 0xFFFFFFFF;

static void MSMLinuxAddCDevice(int devno, MSM_DRIVER_INSTANCE *inst);
static void MSMLinuxDelCDevice(struct cdev *dev_handle);

int mtxserv_open(struct inode *inode, struct file *filp);
int mtxserv_close(struct inode *inode, struct file *filp);
long mtxserv_unlocked_ioctl (struct file *filp, unsigned int cmd, unsigned long arg);
int mtxserv_mmap(struct file *filp, struct vm_area_struct *vma);

void mtxserv_init_proc(void);
void mtxserv_remove_proc(void);
static ssize_t mtxserv_dbglevel_write_proc(struct file *file, const char __user *buffer, size_t count, loff_t *data);
static int mtxserv_dbglevel_show(struct seq_file *sfile, void *iterator);
static int mtxserv_dbglevel_open(struct inode *inode, struct file *file);
static int mtxserv_builddate_show(struct seq_file *sfile, void *iterator);
static int mtxserv_builddate_open(struct inode *inode, struct file *file);
static int mtxserv_ASPMStatus_show(struct seq_file *sfile, void *iterator);
static int mtxserv_ASPMStatus_open(struct inode *inode, struct file *file);

struct file_operations mtxserv_fops={
   .owner   = THIS_MODULE,
   .unlocked_ioctl = mtxserv_unlocked_ioctl,
   .open    = mtxserv_open,
   .release = mtxserv_close,
};

static const struct file_operations mtxserv_builddate_ops = {
   .owner   = THIS_MODULE,
   .open    = mtxserv_builddate_open,
   .read    = seq_read,
   .llseek  = seq_lseek,
   .release = single_release
};

static const struct file_operations mtxserv_ASPMStatus_ops = {
   .owner   = THIS_MODULE,
   .open    = mtxserv_ASPMStatus_open,
   .read    = seq_read,
   .llseek  = seq_lseek,
   .release = single_release
};

static const struct file_operations mtxserv_dbglevel_ops = {
   .owner   = THIS_MODULE,
   .open    = mtxserv_dbglevel_open,
   .read    = seq_read,
   .write   = mtxserv_dbglevel_write_proc,
   .llseek  = seq_lseek,
   .release = single_release
};


//--------------------------------------------------------------------------------
//
// Name :
//
//    MSMLinuxReturnNoSys
//
// Routine Description:
//
//    Return the OS error No System
//
// Arguments:
//
// Return Value:
//
//    the error value
//
//--------------------------------------------------------------------------------
LINUX_DRV_STATUS MSMLinuxReturnNoSys(void)
{
   return ENOSYS;
}

//--------------------------------------------------------------------------------
//
// Name :
//
//    mtxserv_open
//
// Routine Description:
//
//    This function will open an handle to this driver
//
// Arguments:
//
//    inode:   the inode of the driver
//    filp:    pointer to the new driver handle file structure
//
// Return Value:
//
//    OS status code
//
//--------------------------------------------------------------------------------
int mtxserv_open(struct inode *inode, struct file *filp)
{
   MSM_DRIVER_INSTANCE* msmDrvInst = container_of(inode->i_cdev, MSM_DRIVER_INSTANCE, cdev);
   filp->private_data = msmDrvInst;
   return 0;
}

//--------------------------------------------------------------------------------
//
// Name :
//
//    MtxServManagerClose
//
// Routine Description:
//
//    this function will close the driver hanlde
//
// Arguments:
//
//    inode:   the inode of the driver
//    filp:    pointer to the closing driver handle file structure
//
// Return Value:
//
//    OS status code
//
//--------------------------------------------------------------------------------
int mtxserv_close(struct inode *inode, struct file *filp)
{
   return 0;
}

//--------------------------------------------------------------------------------
//
// Name :
//
//    mtxserv_unlocked_ioctl
//
// Routine Description:
//
//    This function will call the command dispatcher with OS specific data
//
// Arguments:
//
//    filp:    pointer to the driver handle file structure
//    cmd:     the command to execute
//    arg:     user data
//
// Return Value:
//
//    OS status code
//
//--------------------------------------------------------------------------------
long mtxserv_unlocked_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
   MSM_DRIVER_INSTANCE* msmDrvInst = filp->private_data;
   return MSMLinuxIoCtl(cmd, (void*)arg, msmDrvInst->msm_device_object, MSM_IOCTL_FROM_USER);
}

//--------------------------------------------------------------------------------
//
// Name :
//
//    MtxServKernelIoctl
//
// Routine Description:
//
//    This function will call the command dispatcher with OS specific data.
//    This function is called by the API when compiled in kernel mode.
//
// Arguments:
//
//    cmd:     the command to execute
//    msg:     user data
//
// Return Value:
//
//    OS status code
//
//--------------------------------------------------------------------------------
int MtxServKernelIoctl (int cmd, void *msg)
{
   return MSMLinuxIoCtl(cmd, (void*)msg, MSMGeneralDevice->msm_device_object, MSM_IOCTL_FROM_KERNEL);
}


//--------------------------------------------------------------------------------
//
// Name :
//
//    MtxServCleanup
//
// Routine Description:
//
//    This function is used when unloading the driver
//
// Arguments:
//
// Return Value:
//
//--------------------------------------------------------------------------------
void MtxServCleanup (void)
{
   mtxserv_remove_proc();
   MSMLinuxDelCDevice(&MSMGeneralDevice->cdev);
   MSMLinuxInstanceCleanup(MSMGeneralDevice->msm_device_object);

   MDrv_memfree(MSMGeneralDevice);
   MSMGeneralDevice = NULL;

   unregister_chrdev_region(DevId, MAX_DEVICE_OBJECT);
}

//--------------------------------------------------------------------------------
//
// Name :
//
//    MtxServInit
//
// Routine Description:
//
//    This function executes at the driver startup
//
// Arguments:
//
// Return Value:
//
//    OS status code
//
//--------------------------------------------------------------------------------
int MtxServInit (void)
{
	if((alloc_chrdev_region(&DevId, 0, MAX_DEVICE_OBJECT, "mtxservmanager")) < 0)
	{
      DbgError ("Can't get a major number\n");
		return -ENODEV;
	}

	MSMGeneralDevice = MDrv_memalloc(sizeof(MSM_DRIVER_INSTANCE));
	if(MSMGeneralDevice == NULL)
	{
		DbgError ("not enougth memory\n");
		return -ENOMEM;
	}
	MDrv_memset(MSMGeneralDevice, 0, sizeof(MSM_DRIVER_INSTANCE));

   MSMLinuxCreateDefaultObject(&MSMGeneralDevice->msm_device_object);
	MSMLinuxAddCDevice(0, MSMGeneralDevice);

   mtxserv_init_proc();
   return 0;
}

static struct proc_dir_entry *MtxServProc = NULL;
//--------------------------------------------------------------------------------
//
// Name :
//
//    mtxserv_init_proc
//
// Routine Description:
//
//    This function initializes the proc callbacks
//
// Arguments:
//
// Return Value:
//
//--------------------------------------------------------------------------------
void mtxserv_init_proc(void)
{
   struct proc_dir_entry *entry;

   MtxServProc = proc_mkdir("driver/mil/mtxservmanager", NULL);

   entry = proc_create_data("dbg_level", 0666, MtxServProc, &mtxserv_dbglevel_ops, NULL);
   entry = proc_create_data("build_date", 0444, MtxServProc, &mtxserv_builddate_ops, NULL);
   entry = proc_create_data("aspm_status", 0444, MtxServProc, &mtxserv_ASPMStatus_ops, NULL);
}

//--------------------------------------------------------------------------------
//
// Name :
//
//    mtxserv_remove_proc
//
// Routine Description:
//
//    This function removes the proc callbacks
//
// Arguments:
//
// Return Value:
//
//--------------------------------------------------------------------------------
void mtxserv_remove_proc(void)
{
   remove_proc_entry("dbg_level", MtxServProc);
   remove_proc_entry("build_date", MtxServProc);
   remove_proc_entry("aspm_status", MtxServProc);
   remove_proc_entry("driver/mil/mtxservmanager", NULL);
}

//--------------------------------------------------------------------------------
//
// Name :
//
//    mtxserv_dbgLevel_read_proc
//
// Routine Description:
//
//    This function executes a read operation on the debug level proc entry
//
// Arguments:
//
// Return Value:
//
//    number of character written
//
//--------------------------------------------------------------------------------
static int mtxserv_dbglevel_show(struct seq_file *sfile, void *iterator)
{
   seq_printf(sfile, "0x%x\n", MDrv_DbgLevel);
   return 0;
}

static int mtxserv_dbglevel_open(struct inode *inode, struct file *file)
{
   return single_open(file, mtxserv_dbglevel_show, NULL);
}

static ssize_t mtxserv_dbglevel_write_proc(struct file *file, const char __user *buffer, size_t count, loff_t *data)
{
   char *localBuf = NULL, *tmp = NULL;
   int newcount = count;
   int val = 0;
   int ret = 0;

   localBuf = MDrv_memalloc(count+1);
   if(localBuf == NULL)
      return -ENOMEM;

   tmp = localBuf;

   while((ret = copy_from_user(tmp, buffer, newcount)))
   {
      tmp += newcount - ret;
      buffer += newcount - ret;
      newcount -= newcount - ret;
   }

   localBuf[count] = '\0';

   sscanf(localBuf, "%x\n", &val);

   MDrv_DbgLevel = val;

   MDrv_memfree(localBuf);

   return count;
}

//--------------------------------------------------------------------------------
//
// Name :
//
//    MtxServManagerReadDriverDate
//
// Routine Description:
//
//    This function executes a read operation on the driver date proc entry
//
// Arguments:
//
// Return Value:
//
//--------------------------------------------------------------------------------
static int mtxserv_builddate_show(struct seq_file *sfile, void *iterator)
   {
   seq_printf(sfile, "Module built on %s at %s\n",  __DATE__, __TIME__);
   return 0;
   }
static int mtxserv_builddate_open(struct inode *inode, struct file *file)
   {
   return single_open(file, mtxserv_builddate_show, NULL);
   }

//--------------------------------------------------------------------------------
//
// Name :
//
//    mtxserv_ASPMStatus_read_proc
//
// Routine Description:
//
//    This function executes a read operation on the driver ASPM state proc entry
//
// Arguments:
//
// Return Value:
//
//--------------------------------------------------------------------------------
static int mtxserv_ASPMStatus_show(struct seq_file *sfile, void *iterator)
   {
   seq_printf(sfile,"ASPM status:\nDone: %u\nFailed: %u\nSeen: %u\n",  GlobalASPMDone, GlobalASPMFailed, GlobalASPMSeen);
   return 0;
   }
static int mtxserv_ASPMStatus_open(struct inode *inode, struct file *file)
   {
   return single_open(file, mtxserv_ASPMStatus_show, NULL);
   }
//--------------------------------------------------------------------------------
//
// Name :
//
//    MSMLinuxAddCDevice
//
// Routine Description:
//
//    This function attach the driver callbacks to the kernel
//
// Arguments:
//
//    devno:   the driver number
//    inst:    pointer to the driver internal structure
//
// Return Value:
//
//--------------------------------------------------------------------------------
void MSMLinuxAddCDevice(int devno, MSM_DRIVER_INSTANCE *inst)
{
	devno = MKDEV(MAJOR(DevId),MINOR(DevId) + devno);
	cdev_init(&inst->cdev, &mtxserv_fops);

	inst->cdev.owner = THIS_MODULE;
	inst->cdev.ops   = &mtxserv_fops;

	if (cdev_add(&inst->cdev, devno, 1) < 0)
	{
		DbgError("Failed to add cdev for: mtxservmanager\n");
	}
}

//--------------------------------------------------------------------------------
//
// Name :
//
//    MSMLinuxDelCDevice
//
// Routine Description:
//
//    This function detach the driver callbacks to the kernel
//
// Arguments:
//
//    dev_handle: pointer to detach
//
// Return Value:
//
//--------------------------------------------------------------------------------
void MSMLinuxDelCDevice(struct cdev *dev_handle)
{
	cdev_del(dev_handle);
}

int MSMLinuxGetSize(unsigned long BarAddress, unsigned long* BarSize)
{
   struct pci_dev* PciDevicePtr = NULL;
   unsigned int iter = 0;

   for_each_pci_dev(PciDevicePtr)
   {
      for(iter = 0; iter < DEVICE_COUNT_RESOURCE; iter++)
      {
         if((pci_resource_flags(PciDevicePtr, iter) & IORESOURCE_MEM) &&
            (pci_resource_start(PciDevicePtr, iter) == (unsigned long)BarAddress))
            {
               *BarSize = pci_resource_len(PciDevicePtr, iter);
               return 0;
            }
      }

   }

   return -1; // no device found
}

void MSMLinuxSetASPMData(unsigned int ASPMDone, unsigned int ASPMFailed, unsigned int ASPMSeen)
   {
   GlobalASPMDone = ASPMDone;
   GlobalASPMFailed = ASPMFailed;
   GlobalASPMSeen = ASPMSeen;

   }


EXPORT_SYMBOL (MtxServKernelIoctl);
EXPORT_SYMBOL (MSMAttach);
EXPORT_SYMBOL (MSMDetach);
EXPORT_SYMBOL (MSMReadConfig);
EXPORT_SYMBOL (MSMWriteConfig);
EXPORT_SYMBOL (MSMFindDevice);
EXPORT_SYMBOL (MSMFindDeviceEx);
EXPORT_SYMBOL (MSMRescanPCI);
EXPORT_SYMBOL (MSMGetBarSize);
EXPORT_SYMBOL (MSMGetBarSizeFromRegistry);
EXPORT_SYMBOL (MSMGetLicenseInfo);
EXPORT_SYMBOL (MSMAddLicenseInfo);
EXPORT_SYMBOL (MSMGetDriverInfo);
EXPORT_SYMBOL (MSMGetPciDeviceArray);

module_init(MtxServInit);
module_exit(MtxServCleanup);
